//imports

import java.awt.BorderLayout;
import java.awt.Container;
import java.awt.FlowLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.image.BufferedImage;
import java.awt.image.DataBuffer;
import java.awt.image.DataBufferByte;
import java.awt.image.WritableRaster;
import java.io.File;
import java.io.IOException;
import javax.imageio.ImageIO;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JOptionPane;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JSpinner;
import javax.swing.SpinnerNumberModel;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.UIManager;

/**
 * Graphics - A graphics coursework in Java.
 *
 * I declare this work is mine in its entirety with the exception
 * of template code provided by Dr. Mark Jones of Swansea University.
 * 
 * @author Sean Handley (320097@swan.ac.uk)
 * @version 2006-02
 * 
 */
public class Graphics extends JFrame {
	static final long serialVersionUID = 1L;
	JButton load_button, invert_button, slow_gamma_button, fast_gamma_button, convolve_button;
	JSpinner intensity_spinner;
	SpinnerNumberModel numModel;
	
	JLabel image_icon;
	BufferedImage image;
	double gamma;
	
	private BufferedImage loadImage(String filename) {
		File image_file = new File(filename);
		
		try {
			// Open the file and read it into a BufferedImage
			image = ImageIO.read(image_file);
		}
		catch(IOException e) {
			e.printStackTrace();
		}
		return image;
	}
	/*
	 This function sets up the GUI
	 */
	public Graphics() throws IOException {
		
		//set up the superclass (JFrame)
		super();
		
		//use raytrace as default image
		loadImage("raytrace.jpg");
		
		// Set up the simple GUI
		// First the container:
		Container container = getContentPane();
		container.setLayout(new BorderLayout());
		
		//Set up the button container:
		Container buttons = new Container();
		buttons.setLayout(new FlowLayout());
		
		//Setup a container for the spinner
		Container spinnerCont = new Container();
		spinnerCont.setLayout(new FlowLayout());
		
		//setup the load button
		load_button = new JButton("Load...");
		buttons.add(load_button);
		
		//add the gamma label
		JLabel intensityLabel = new JLabel("Gamma (0-255)");
		spinnerCont.add(intensityLabel);
		
		//set up the spinner
		numModel = new SpinnerNumberModel(1,0,255,0.01);
		intensity_spinner = new JSpinner(numModel);
		spinnerCont.add(intensity_spinner);
		
		//add the spinner & label to the buttons container
		buttons.add(spinnerCont);
		
		//make a new container for the gamma buttons
		Container gammaButtons = new Container();
		gammaButtons.setLayout(new FlowLayout());
		
		// Setup the Gamma button
		slow_gamma_button = new JButton("Slow Gamma");
		gammaButtons.add(slow_gamma_button);
		fast_gamma_button = new JButton("Fast Gamma");
		gammaButtons.add(fast_gamma_button);
		
		//make a container for the convolve button
		Container convolveCont = new Container();
		convolveCont.setLayout(new FlowLayout());
		
		// Convolve button
		convolve_button = new JButton("Apply Filter...");
		
		//add the gamma buttons to the buttons container
		buttons.add(gammaButtons);
		
		//make a container for the invert button
		Container invertCont = new Container();
		invertCont.setLayout(new FlowLayout());
		
		// Then the invert button
		invert_button = new JButton("Invert");
		
		//add the invert button to its container
		//and put it in the buttons container
		invertCont.add(invert_button);
		buttons.add(invertCont);
		
		//add the convolve button to its container
		//and put it in the buttons container
		
		convolveCont.add(convolve_button);
		buttons.add(convolveCont);
		
		// Then our image (as a label icon) at the top
		image_icon=new JLabel(new ImageIcon(image));
		container.add(image_icon,BorderLayout.NORTH);
		
		//add the buttons container at the bottom
		container.add(buttons,BorderLayout.SOUTH);
		
		// Now all the handlers
		GUIActionHandler action = new GUIActionHandler();
		GUIChangeHandler change = new GUIChangeHandler();
		
		// Button handlers
		load_button.addActionListener(action);
		invert_button.addActionListener(action);
		slow_gamma_button.addActionListener(action);
		fast_gamma_button.addActionListener(action);
		convolve_button.addActionListener(action);
		intensity_spinner.addChangeListener(change);
		
		// ...and display everything
		pack();
		setLocationRelativeTo(null);
		setTitle("Graphics Coursework I: Sean Handley  (320097@swan.ac.uk)");
		setVisible(true);
		
	}
	private class GUIChangeHandler implements ChangeListener {
		public void stateChanged(ChangeEvent arg0) {
			gamma = ((Double) intensity_spinner.getValue()).doubleValue();
		}
	}
	/*
	 This is the event handler for the application
	 */
	private class GUIActionHandler implements ActionListener {
		public void actionPerformed(ActionEvent event) { 
			
			if(event.getSource()==load_button) {
				//load the image
				String filename = "raytrace.jpg";
				JFileChooser fc = new JFileChooser(new File(filename));
				
				// Show open dialog; this method does not return until the dialog is closed
				fc.showOpenDialog(rootPane);
				try {
					String selFile = fc.getSelectedFile().toString();
					image_icon.setIcon(new ImageIcon(loadImage(selFile)));
				}
				catch(NullPointerException e) {
					//do nothing
				}
			} else if (event.getSource()==invert_button) {
				// Call image processing function
				image=Invert(image);
				// Update image
				image_icon.setIcon(new ImageIcon(image));
			} else if (event.getSource()==slow_gamma_button) {
				// Call image processing function
				image=SlowGamma(image);
				// Update image
				image_icon.setIcon(new ImageIcon(image));
			} else if (event.getSource()==fast_gamma_button) {
				// Call image processing function
				image=FastGamma(image);
				// Update image
				image_icon.setIcon(new ImageIcon(image));
			} else if (event.getSource()==convolve_button) {
				FilterDiag fd = new FilterDiag();
				int filter[][] = fd.makeDiag();
				
				try {
					if(checkFilter(filter)) {
						// Call image processing function
						image=applyFilter(image,filter);
						// Update image
						image_icon.setIcon(new ImageIcon(image));
					} else {
						errMessage("The filter you have selected is invalid. Filters must be NxN or NxM quadrlaterals.","Invalid Filter");
					}
				}
				catch(ArrayIndexOutOfBoundsException e) {
					//do nothing, user has probably X'd the dialog
				}
				catch(NullPointerException e) {
					//do nothing, user has probably X'd the dialog
				}
				catch(Exception e) {
					e.printStackTrace();
				}
			}
		}
	}
	
	public void errMessage(String message, String title) {
		JOptionPane.showMessageDialog(this,message,title,
				JOptionPane.ERROR_MESSAGE);
	}
	
	/*
	 This function will return a pointer to an array
	 of bytes which represent the image data in memory.
	 Using such a pointer allows fast access to the image
	 data for processing (rather than getting/setting
	 individual pixels)
	 */
	public static byte[] GetImageData(BufferedImage image) {
		WritableRaster WR=image.getRaster();
		DataBuffer DB=WR.getDataBuffer();
		if (DB.getDataType() != DataBuffer.TYPE_BYTE)
			throw new IllegalStateException("That's not of type byte");
		
		return ((DataBufferByte) DB).getData();
	}
	
	/*
	 This function shows how to carry out an operation on an image.
	 It obtains the dimensions of the image, and then loops through
	 the image carrying out the invert. It also shows a faster version
	 commented out
	 */
	public BufferedImage Invert(BufferedImage image) {
		//Get image dimensions, and declare loop variables
		int w=image.getWidth(), h=image.getHeight(), i, j, c;
		//Obtain pointer to data for fast processing
		byte[] data = GetImageData(image);
		
		//Shows how to loop through each pixel and colour
		//Try to always use j for loops in y, and i for loops in x
		//as this makes the code more readable
		for (j=0; j<h; j++) {
			for (i=0; i<w; i++) {
				for (c=0; c<3; c++) {
					data[c+3*i+3*j*w]=(byte) (255-(data[c+3*i+3*j*w]&255));
				} // colour loop
			} // column loop
		} // row loop
		
		//Faster version - array index requires no calculation
		/*
		 for (j=0; j<h*w*3; j++) {
		 data[j]=(byte) (255-(data[j]&255));
		 }
		 */
		return image;
	}
	
	public BufferedImage SlowGamma(BufferedImage image) {
		//Get image dimensions, and declare loop variables
		int w=image.getWidth(), h=image.getHeight(), i, j, c;
		//Obtain pointer to data for fast processing
		byte[] data = GetImageData(image);
		
		for (j=0; j<h; j++) {
			for (i=0; i<w; i++) {
				for (c=0; c<3; c++) {
					data[c+3*i+3*j*w]= (new Double(Math.pow(((data[c+3*i+3*j*w]&255)/256.0),gamma)*256.0)).byteValue();
				} // colour loop
			} // column loop
		} // row loop
		
		return image;
	}
	
	public BufferedImage FastGamma(BufferedImage image) {
		//Get image dimensions, and declare loop variables
		int w=image.getWidth(), h=image.getHeight(), j, i;
		
		//Obtain pointer to data for fast processing
		byte[] data = GetImageData(image);
	
		//set up array for values
		int vals[] = new int[256];

		for (i=0; i < 256; i++) {
			vals[i] = new Double(Math.pow((i/256.0),gamma) * 256.0).intValue();
		}
	
		for (j=0; j<h*w*3; j++) {
			data[j]= (byte) vals[(data[j]&255)];
		}
		return image;
	}
	
	public BufferedImage BlueFade(BufferedImage image) {
		//Get image dimensions, and declare loop variables
		int w=image.getWidth(), h=image.getHeight(), i, j, c;
		
		//Obtain pointer to data for fast processing
		byte[] data = GetImageData(image);
		
		int int_image[][][];
		
		int_image = new int[h][w][3];
		
		// Copy byte data to new image taking care to treat bytes as unsigned
		for (j=0; j<h; j++) {
			for (i=0; i<w; i++) {
				for (c=0; c<3; c++) {
					int_image[j][i][c]=data[c+3*i+3*j*w]&255;
				} // colour loop
			} // column loop
		} // row loop
		
		// Now carry out processing on this different data typed image (e.g. convolution or "bluefade"
		for (j=0; j<h; j++) {
			for (i=0; i<w; i++) {
				int_image[j][i][0]=255*j/h; //BLUE
				int_image[j][i][1]=0; //GREEN
				int_image[j][i][2]=0; //RED
			} // column loop
		} // row loop
		
		// Now copy the processed image back to the original
		for (j=0; j<h; j++) {
			for (i=0; i<w; i++) {
				for (c=0; c<3; c++) {
					data[c+3*i+3*j*w]=(byte) int_image[j][i][c];
				} // colour loop
			} // column loop
		} // row loop
		
		
		return image;
	}
	/*
	 * Check that each row in the filter is the same length
	 * (and, therefore, a quadrilateral)
	 */
	public boolean checkFilter(int filter[][]) {
		boolean valid = true;
		int i = 0;
		
		
		//all rows should match the length of the first row
		int sizeOfRow = (filter[i]).length;
		
		for(i=0; i < filter.length; i++) {
			if((filter[i]).length != sizeOfRow) {
				valid = false;
				break;
			}
		}
		
		//catch the empty filter
		if(sizeOfRow == 0) { valid = false;}
		
		return valid;
	}
	
	public BufferedImage applyFilter(BufferedImage image, int filter[][]) {
		//Get image dimensions, and declare loop variables
		int w=image.getWidth(), h=image.getHeight(), i, j, c;
		
		//Obtain pointer to data for fast processing
		byte[] data = GetImageData(image);
		
		int rowgap = new Double(Math.ceil(filter[0].length / 2.0)).intValue();
		int colgap = new Double(Math.ceil(filter.length / 2.0)).intValue();
		
		//reduce by one if even to balance ceil function
		if((filter[0].length % 2) != 0) {
			rowgap--;
		}

		if((filter.length %2) != 0) {
			colgap--;
		}
	
		//System.out.println("filter[0]: " + filter[0][3]);
		//System.out.println("rowgap: " + rowgap + ", colgap: " + colgap);
		
		//max 
		int maxb = 0, maxg = 0, maxr = 0;
		int minb = 0, ming = 0, minr = 0;
		
		int int_image[][][];
		int new_image[][][];
		
		int_image = new int[h][w][3];
		new_image = new int[h][w][3];
		
		// Copy byte data to new image taking care to treat bytes as unsigned
		for (j=0; j<h; j++) {
			for (i=0; i<w; i++) {
				for (c=0; c<3; c++) {
					int_image[j][i][c]=data[c+3*i+3*j*w]&255;
				} // colour loop
			} // column loop
		} // row loop
		
		// Now carry out processing on this different data typed image
		for (j=0; j<h; j++) {
			for (i=0; i<w; i++) {
				//if central pixel of filter is valid...
				if(((i - rowgap) >= 0)
					&& ((j - colgap) >= 0)
					&& ((i + rowgap) < w)
					&& ((j + colgap) < h)) {
					
					//get top-left coord (starting position)
					
					int startj = j - colgap;
					int starti = i - rowgap;
					
					//loop through filter
					for(int x = 0; x < filter.length; x++) {
						for(int y = 0; y < filter[0].length; y++) {
							//accumulate values
							new_image[j][i][0] += (int_image[startj + x][starti + y][0]) * filter[x][y]; //BLUE
							new_image[j][i][1] += (int_image[startj + x][starti + y][1]) * filter[x][y];//GREEN
							new_image[j][i][2] += (int_image[startj + x][starti + y][2]) * filter[x][y];//RED
						}
					}		
				}
			} // row loop
		} // column loop
		
		//loop through image
		for(int m = 0; m < h; m++) {
			for(int n = 0; n < w; n++) {
				//get max values
				if(new_image[m][n][0] > maxb) { maxb = new_image[m][n][0];}
				if(new_image[m][n][1] > maxg) { maxg = new_image[m][n][1];}
				if(new_image[m][n][2] > maxr) { maxr = new_image[m][n][2];}
			}
		}
		
		//set min to a number that will have less thans
		minb = maxb;
		ming = maxg;
		minr = maxr;
		
		//loop through image
		for(int m = 0; m < h; m++) {
			for(int n = 0; n < w; n++) {
				//get max values
				if(new_image[m][n][0] < minb) { minb = new_image[m][n][0];}
				if(new_image[m][n][1] < ming) { ming = new_image[m][n][1];}
				if(new_image[m][n][2] < minr) { minr = new_image[m][n][2];}
			}
		}
		
		//fix the top edge
		for(j = 0; j < colgap; j++) {
			for(i = 0; i < w; i++) {
				new_image[j][i][0] = new_image[j+colgap][i][0];
				new_image[j][i][1] = new_image[j+colgap][i][1];
				new_image[j][i][2] = new_image[j+colgap][i][2];
			}
		}
		
		//fix the bottom edge
		for(j = (h-1); j > (h-colgap-1); j--) {
			for(i = 0; i < w; i++) {
				new_image[j][i][0] = new_image[j-colgap][i][0];
				new_image[j][i][1] = new_image[j-colgap][i][1];
				new_image[j][i][2] = new_image[j-colgap][i][2];
			}
		}
		
		//fix the left edge
		for(i = 0; i < rowgap; i++) {
			for(j = 0; j < h; j++) {
				new_image[j][i][0] = new_image[j][i+rowgap][0];
				new_image[j][i][1] = new_image[j][i+rowgap][1];
				new_image[j][i][2] = new_image[j][i+rowgap][2];
			}
		}
		
		//fix the right edge
		for(i = (w-1); i > (w-rowgap-1); i--) {
			for(j = 0; j < h; j++) {
				new_image[j][i][0] = new_image[j][i-rowgap][0];
				new_image[j][i][1] = new_image[j][i-rowgap][1];
				new_image[j][i][2] = new_image[j][i-rowgap][2];
			}
		}
		
		for(j = 0; j < h; j++) {
			for(i=0; i < w; i++) {
				//now for normalisation
				int_image[j][i][0] = ((new_image[j][i][0] - minb) * 255) / (maxb - minb);
				int_image[j][i][1] = ((new_image[j][i][1] - ming) * 255) / (maxg - ming);
				int_image[j][i][2] = ((new_image[j][i][2] - minr) * 255) / (maxr - minr);
			}
		}
		
		// Now copy the processed image back to the original
		for (j=0; j<h; j++) {
			for (i=0; i<w; i++) {
				for (c=0; c<3; c++) {
					data[c+3*i+3*j*w]=(byte) int_image[j][i][c];
				} // colour loop
			} // column loop
		} // row loop
		
		return image;
	}
	
	public static void main(String[] args) throws IOException {
		try {
			UIManager.setLookAndFeel(
			UIManager.getSystemLookAndFeelClassName());
		}
		catch (Exception e) { }

		Graphics graphicsInstance = new Graphics();
		graphicsInstance.toFront();
	}
}
